import re
from typing import Dict, Optional, cast

from django.core.exceptions import ImproperlyConfigured
from django.urls import reverse

from django_backblaze_b2.options import BackblazeB2StorageOptions, CDNConfig
from django_backblaze_b2.storage import BackblazeB2Storage, logger


class PublicStorage(BackblazeB2Storage):
    """
    Storage that requires no authentication to view.
    If the bucket is public, returns the bucket's url, or CDN if configured
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._bucketType: Optional[bool] = None
        self._cdnConfig: Optional[CDNConfig] = kwargs.get("opts", {}).get("cdnConfig")

    def _isPublicBucket(self) -> bool:
        if self._bucketType is None:
            self._bucketType = self.bucket.as_dict().get("bucketType") == "allPublic"
        return self._bucketType

    def _getFileUrl(self, name: str) -> str:
        if not self._isPublicBucket():
            return reverse("django_b2_storage:b2-public", args=[name])
        if self._cdnConfig:
            fileUrl = self.getBackblazeUrl(name)
            cdnUrlBase = self._cdnConfig["baseUrl"].replace("https://", "").replace("http://", "").strip("/")
            if self._cdnConfig["includeBucketUrlSegments"]:
                return re.sub(r"f\d+\.backblazeb2\.com", cdnUrlBase, fileUrl)
            return re.sub(r"f\d+\.backblazeb2\.com/file/[^/]+/", cdnUrlBase + "/", fileUrl)
        return self.getBackblazeUrl(name)

    def _getDjangoSettingsOptions(self, kwargOpts: Dict) -> BackblazeB2StorageOptions:
        _validateKwargOpts(kwargOpts)
        if kwargOpts.get("cdnConfig"):
            if not isinstance(kwargOpts["cdnConfig"], dict):
                raise ImproperlyConfigured("django-backblaze-b2 cdnConfig must be a dict")
            if not isinstance(kwargOpts["cdnConfig"].get("baseUrl"), str):
                raise ImproperlyConfigured("cdnConfig.baseUrl must be a string")
            if not isinstance(kwargOpts["cdnConfig"].get("includeBucketUrlSegments"), bool):
                logger.debug("will treat cdnConfig.includeBucketUrlSegments to False")
        options = super()._getDjangoSettingsOptions(kwargOpts)
        _adjustOptions(options, specificBucket="public")
        return options


class LoggedInStorage(BackblazeB2Storage):
    """Storage that requires authentication to view or download files"""

    def _getFileUrl(self, name: str) -> str:
        return reverse("django_b2_storage:b2-logged-in", args=[name])

    def _getDjangoSettingsOptions(self, kwargOpts: Dict) -> BackblazeB2StorageOptions:
        _validateKwargOpts(kwargOpts)
        options = super()._getDjangoSettingsOptions(kwargOpts)
        _adjustOptions(options, specificBucket="loggedIn")
        return options


class StaffStorage(BackblazeB2Storage):
    """Storage that requires staff permission to view or download files"""

    def _getFileUrl(self, name: str) -> str:
        return reverse("django_b2_storage:b2-staff", args=[name])

    def _getDjangoSettingsOptions(self, kwargOpts: Dict) -> BackblazeB2StorageOptions:
        _validateKwargOpts(kwargOpts)
        options = super()._getDjangoSettingsOptions(kwargOpts)
        _adjustOptions(options, specificBucket="staff")
        return options


def _validateKwargOpts(kwargOpts) -> None:
    if "bucket" in kwargOpts:
        raise ImproperlyConfigured("May not specify 'bucket' in proxied storage class")
    if "application_key_id" in kwargOpts or "application_key" in kwargOpts or "realm" in kwargOpts:
        raise ImproperlyConfigured("May not specify auth credentials in proxied storage class")


def _adjustOptions(options: BackblazeB2StorageOptions, specificBucket: str) -> None:
    proxiedStorageBucketName = options.get("specificBucketNames", {}).get(specificBucket)
    if proxiedStorageBucketName:
        options["bucket"] = cast(str, proxiedStorageBucketName)



from datetime import datetime
from hashlib import sha3_224 as hash
from logging import getLogger
from typing import IO, Any, Callable, Dict, List, Optional, Tuple, cast

from b2sdk.account_info import InMemoryAccountInfo
from b2sdk.account_info.abstract import AbstractAccountInfo
from b2sdk.account_info.sqlite_account_info import SqliteAccountInfo
from b2sdk.api import B2Api, Bucket
from b2sdk.cache import AuthInfoCache
from b2sdk.exception import FileOrBucketNotFound, NonExistentBucket
from django.core.cache.backends.base import BaseCache
from django.core.exceptions import ImproperlyConfigured
from django.core.files.base import File
from django.core.files.storage import Storage
from django.utils.deconstruct import deconstructible
from typing_extensions import TypedDict

from django_backblaze_b2.b2_file import B2File
from django_backblaze_b2.cache_account_info import DjangoCacheAccountInfo
from django_backblaze_b2.options import (
    BackblazeB2StorageOptions,
    DjangoCacheAccountInfoConfig,
    SqliteAccountInfoConfig,
    getDefaultB2StorageOptions,
)

logger = getLogger("django-backblaze-b2")


class _BaseFileInfoDict(TypedDict):
    fileId: str
    fileName: str
    fileInfo: dict


class _FileInfoDict(_BaseFileInfoDict, total=False):
    size: int
    uploadTimestamp: int
    contentType: str


class B2FileInformationNotAvailableException(Exception):
    ...


@deconstructible
class BackblazeB2Storage(Storage):
    """Storage class which fulfills the Django Storage contract through b2 apis"""

    def __init__(self, **kwargs):
        opts = self._getDjangoSettingsOptions(kwargs.get("opts", {}))
        if "opts" in kwargs:
            self._validateOptions(kwargs.get("opts"))
        _merge(opts, kwargs.get("opts", {}))
        logOpts = opts.copy()
        logOpts.update({"application_key_id": "<redacted>", "application_key": "<redacted>"})
        logger.debug(f"Initializing {self.__class__.__name__} with options {logOpts}")

        self._bucketName = opts["bucket"]
        self._defaultFileMetadata = opts["defaultFileInfo"]
        self._forbidFilePropertyCaching = opts["forbidFilePropertyCaching"]
        self._authInfo = dict(
            [(k, v) for k, v in opts.items() if k in ["realm", "application_key_id", "application_key"]]
        )
        self._allowFileOverwrites = opts["allowFileOverwrites"]

        self._getAccountInfo = self._createAccountInfoCallable(opts)

        logger.info(f"{self.__class__.__name__} instantiated to use bucket {self._bucketName}")
        if opts["authorizeOnInit"]:
            logger.debug(f"{self.__class__.__name__} authorizing")
            self.b2Api
            if opts["validateOnInit"]:
                self._getOrCreateBucket(opts["nonExistentBucketDetails"])

    def _getDjangoSettingsOptions(self, kwargOpts: Dict) -> BackblazeB2StorageOptions:
        """Setting terminology taken from:
        https://b2-sdk-python.readthedocs.io/en/master/glossary.html#term-application-key-ID
        kwargOpts available for subclasses
        """
        from django.conf import settings

        if not hasattr(settings, "BACKBLAZE_CONFIG"):
            raise ImproperlyConfigured("add BACKBLAZE_CONFIG dict to django settings")
        if "application_key_id" not in settings.BACKBLAZE_CONFIG or "application_key" not in settings.BACKBLAZE_CONFIG:
            raise ImproperlyConfigured(
                "At minimum BACKBLAZE_CONFIG must contain auth 'application_key' and 'application_key_id'"
                f"\nfound: {settings.BACKBLAZE_CONFIG}"
            )
        self._validateOptions(settings.BACKBLAZE_CONFIG)
        opts = getDefaultB2StorageOptions()
        opts.update(settings.BACKBLAZE_CONFIG)  # type: ignore
        return opts

    def _validateOptions(self, options: Dict) -> None:
        unrecognizedOptions = [k for k in options.keys() if k not in getDefaultB2StorageOptions().keys()]
        if unrecognizedOptions:
            raise ImproperlyConfigured(f"Unrecognized options: {unrecognizedOptions}")

    def _createAccountInfoCallable(self, opts: BackblazeB2StorageOptions) -> Callable[[], AbstractAccountInfo]:
        if (
            not isinstance(opts["accountInfo"], dict)
            or "type" not in opts["accountInfo"]
            or opts["accountInfo"]["type"] not in ["memory", "sqlite", "django-cache"]
        ):
            raise ImproperlyConfigured(
                (f"accountInfo property must be a dict with type found in options.py, was {opts['accountInfo']}")
            )
        if opts["accountInfo"]["type"] == "django-cache":
            logger.debug(f"{self.__class__.__name__} will use {DjangoCacheAccountInfo.__name__}")
            return lambda: DjangoCacheAccountInfo(
                cacheName=cast(DjangoCacheAccountInfoConfig, opts["accountInfo"]).get("cache", "django-backblaze-b2")
            )
        elif opts["accountInfo"]["type"] == "memory":
            logger.debug(f"{self.__class__.__name__} will use {InMemoryAccountInfo.__name__}")
            return lambda: InMemoryAccountInfo()
        elif opts["accountInfo"]["type"] == "sqlite":
            logger.debug(f"{self.__class__.__name__} will use {SqliteAccountInfo.__name__}")
            return lambda: SqliteAccountInfo(
                file_name=cast(SqliteAccountInfoConfig, opts["accountInfo"])["databasePath"]
            )
        raise ImproperlyConfigured()

    @property
    def b2Api(self) -> B2Api:
        if not hasattr(self, "_b2Api"):
            self._accountInfo = self._getAccountInfo()
            self._b2Api = B2Api(account_info=self._accountInfo, cache=AuthInfoCache(self._accountInfo))
            self._b2Api.authorize_account(**self._authInfo)
        return self._b2Api

    @property
    def bucket(self) -> Bucket:
        if not hasattr(self, "_bucket"):
            self._getOrCreateBucket()
        return self._bucket

    def _getOrCreateBucket(self, newBucketDetails=None) -> None:
        try:
            self._bucket = self.b2Api.get_bucket_by_name(self._bucketName)
        except NonExistentBucket as e:
            if newBucketDetails is not None:
                logger.debug(f"Bucket {self._bucketName} not found. Creating with details: {newBucketDetails}")
                if "bucket_type" not in newBucketDetails:
                    newBucketDetails["bucket_type"] = "allPrivate"
                self._bucket = self.b2Api.create_bucket(name=self._bucketName, **newBucketDetails)
            else:
                raise e
        logger.debug(f"Connected to bucket {self._bucket.as_dict()}")

    def _open(self, name: str, mode: str) -> File:
        return B2File(
            name=name, bucket=self.bucket, fileMetadata=self._defaultFileMetadata, mode=mode, sizeProvider=self.size,
        )

    def _save(self, name: str, content: IO[Any]) -> str:
        """
        Save and retrieve the filename.
        If the file exists it will make another version of that file.
        """
        return B2File(
            name=name, bucket=self.bucket, fileMetadata=self._defaultFileMetadata, mode="w", sizeProvider=self.size,
        ).saveAndRetrieveFile(content)

    def path(self, name: str) -> str:
        return name

    def delete(self, name: str) -> None:
        fileInfo = self._fileInfo(name)
        if fileInfo:
            logger.debug(f"Deleting file {name} id=({fileInfo['fileId']})")
            self.b2Api.delete_file_version(file_id=fileInfo["fileId"], file_name=name)
            if self._cache:
                self._cache.delete(self._fileCacheKey(name))
        else:
            logger.debug("Not found")

    def _fileInfo(self, name: str) -> Optional[_FileInfoDict]:
        try:
            if self._cache:
                cacheKey = self._fileCacheKey(name)
                timeoutInSeconds = 60

                def loadInfo():
                    logger.debug(f"file info cache miss for {name}")
                    return self.bucket.get_file_info_by_name(name).as_dict()

                return self._cache.get_or_set(key=cacheKey, default=loadInfo, timeout=timeoutInSeconds)
            return self.bucket.get_file_info_by_name(name).as_dict()
        except FileOrBucketNotFound:
            return None

    def _fileCacheKey(self, name: str) -> str:
        return hash(f"{self.bucket.name}__{name}".encode()).hexdigest()

    @property
    def _cache(self) -> Optional[BaseCache]:
        if (
            not self._forbidFilePropertyCaching
            and self.b2Api  # force init
            and self._accountInfo
            and isinstance(self._accountInfo, DjangoCacheAccountInfo)
        ):
            return self._accountInfo.cache
        return None

    def exists(self, name: str) -> bool:
        return bool(self._fileInfo(name))

    def size(self, name: str) -> int:
        fileInfo = self._fileInfo(name)
        return fileInfo.get("size", 0) if fileInfo else 0

    def url(self, name: Optional[str]) -> str:
        if not name:
            raise Exception("Name must be defined")
        return self._getFileUrl(name)

    def _getFileUrl(self, name: str) -> str:
        return self.getBackblazeUrl(name)

    def getBackblazeUrl(self, filename: str) -> str:
        return self.b2Api.get_download_url_for_file_name(bucket_name=self._bucketName, file_name=filename)

    def get_available_name(self, name: str, max_length: Optional[int] = None) -> str:
        if self._allowFileOverwrites:
            return name
        return super().get_available_name(name, max_length)

    def listdir(self, path: str) -> Tuple[List[str], List[str]]:
        """
        List the contents of the specified path. Return a 2-tuple of lists:
        the first item being directories, the second item being files.
        """
        raise NotImplementedError("subclasses of Storage must provide a listdir() method")

    def get_accessed_time(self, name: str) -> datetime:
        """
        Return the last accessed time (as a datetime) of the file specified by
        name. The datetime will be timezone-aware if USE_TZ=True.
        """
        raise NotImplementedError("subclasses of Storage must provide a get_accessed_time() method")

    def get_created_time(self, name: str) -> datetime:
        """
        Return the creation time (as a datetime) of the file specified by name.
        The datetime will be timezone-aware if USE_TZ=True.
        """
        from datetime import timezone

        from django.conf import settings

        fileInfo = self._fileInfo(name)
        try:
            if fileInfo and float(fileInfo.get("uploadTimestamp", 0)) > 0:
                timestamp = float(fileInfo["uploadTimestamp"]) / 1000.0
                if settings.USE_TZ:
                    # Safe to use .replace() because UTC doesn't have DST
                    return datetime.utcfromtimestamp(timestamp).replace(tzinfo=timezone.utc)
                return datetime.fromtimestamp(timestamp)
        except ValueError as e:
            raise B2FileInformationNotAvailableException(f"'uploadTimestamp' from API not valid for {name}: {e}")
        raise B2FileInformationNotAvailableException(f"'uploadTimestamp' not available for {name}")

    def get_modified_time(self, name: str) -> datetime:
        """
        Return the last modified time (as a datetime) of the file specified by
        name. The datetime will be timezone-aware if USE_TZ=True.
        """
        return self.get_created_time(name)


def _merge(target: Dict, source: Dict, path=None) -> Dict:
    """merges b into a
    https://stackoverflow.com/a/7205107/11076240
    """
    if path is None:
        path = []
    for key in source:
        if key in target:
            printablePath = ".".join(path + [str(key)])
            if isinstance(target[key], dict) and isinstance(source[key], dict):
                _merge(target[key], source[key], path + [str(key)])
            elif target[key] != source[key]:
                logger.debug(f"Overriding setting {printablePath} with value {source[key]}")
                target[key] = source[key]
        else:
            target[key] = source[key]
    return target



from typing import Any, Dict, Optional, Union

from typing_extensions import Literal, TypedDict


class ProxiedBucketNames(TypedDict):
    public: Optional[str]
    loggedIn: Optional[str]
    staff: Optional[str]


class DjangoCacheAccountInfoConfig(TypedDict):
    type: Literal["django-cache"]
    cache: str


class InMemoryAccountInfoConfig(TypedDict):
    type: Literal["memory"]


class SqliteAccountInfoConfig(TypedDict):
    type: Literal["sqlite"]
    databasePath: str


class CDNConfig(TypedDict):
    baseUrl: str
    includeBucketUrlSegments: bool


class BackblazeB2StorageOptions(TypedDict):
    """Configuration options."""

    realm: str  # default "production"
    application_key_id: str
    application_key: str
    bucket: str
    authorizeOnInit: bool
    validateOnInit: bool
    allowFileOverwrites: bool
    accountInfo: Optional[Union[DjangoCacheAccountInfoConfig, InMemoryAccountInfoConfig, SqliteAccountInfoConfig]]
    forbidFilePropertyCaching: bool
    specificBucketNames: ProxiedBucketNames
    cdnConfig: Optional[CDNConfig]
    # see: https://b2-sdk-python.readthedocs.io/en/master/api/api.html#b2sdk.v1.B2Api.create_bucket
    nonExistentBucketDetails: Optional[Dict[str, Union[str, Dict[str, Any]]]]
    defaultFileInfo: Dict[str, Any]


def getDefaultB2StorageOptions() -> BackblazeB2StorageOptions:
    return {
        "realm": "production",
        "application_key_id": "you must set this value yourself",
        "application_key": "you must set this value yourself",
        "bucket": "django",
        "authorizeOnInit": True,
        "validateOnInit": True,
        "allowFileOverwrites": False,
        "accountInfo": {"type": "django-cache", "cache": "django-backblaze-b2"},
        "forbidFilePropertyCaching": False,
        "specificBucketNames": {"public": None, "loggedIn": None, "staff": None},
        "cdnConfig": None,
        "nonExistentBucketDetails": None,
        "defaultFileInfo": {},
    }



import logging
from functools import wraps
from hashlib import sha3_224 as hash
from typing import Iterable, Optional, Tuple

from b2sdk.account_info.exception import MissingAccountData
from b2sdk.account_info.upload_url_pool import UrlPoolAccountInfo
from django.core.cache import InvalidCacheBackendError, caches
from django.core.exceptions import ImproperlyConfigured

logger = logging.getLogger("django-backblaze-b2")


class StoredBucketInfo:
    name: str
    id_: str


def _raise_missing_if_result_is_none(function):
    """
    Raise MissingAccountData if function's result is None.
    adapted from https://github.com/Backblaze/b2-sdk-python/blob/v1.5.0/b2sdk/account_info/in_memory.py
    """

    @wraps(function)
    def getter_function(*args, **kwargs):
        assert function.__name__.startswith("get_")
        result = function(*args, **kwargs)
        if result is None:
            raise MissingAccountData(function.__name__[len("get_") :])
        return result

    return getter_function


class DjangoCacheAccountInfo(UrlPoolAccountInfo):
    """
    Store account information in django's cache: https://docs.djangoproject.com/en/3.1/topics/cache
    """

    def __init__(self, cacheName: str):
        logger.debug(f"Initializing {self.__class__.__name__} with cache '{cacheName}'")
        self._cacheName = cacheName
        try:
            self.cache = caches[cacheName]
            self.cache.set("bucket_names", [])
        except InvalidCacheBackendError:
            logger.exception("Cache assignment failed")
            from django.conf import settings

            helpMessage = (
                (
                    ". "
                    "The default 'accountInfo' option of this library is with a django cache"
                    " by the name of 'django-backblaze-b2'"
                )
                if "accountInfo" not in settings.BACKBLAZE_CONFIG
                else ""
            )

            raise ImproperlyConfigured(f"Expected to find a cache with name '{cacheName}' as per options" + helpMessage)
        super(DjangoCacheAccountInfo, self).__init__()

    def clear(self):
        """
        Remove all info about accounts and buckets.
        """
        self.cache.clear()
        self.cache.set("bucket_names", [])

    def _set_auth_data(
        self,
        account_id,
        auth_token,
        api_url,
        download_url,
        recommended_part_size,
        absolute_minimum_part_size,
        application_key,
        realm,
        s3_api_url,
        allowed,
        application_key_id,
    ):
        self.cache.set("account_id", account_id)
        self.cache.set("auth_token", auth_token)
        self.cache.set("api_url", api_url)
        self.cache.set("download_url", download_url)
        self.cache.set("recommended_part_size", recommended_part_size)
        self.cache.set("absolute_minimum_part_size", absolute_minimum_part_size)
        self.cache.set("application_key", application_key)
        self.cache.set("realm", realm)
        self.cache.set("s3_api_url", s3_api_url)
        self.cache.set("allowed", allowed)
        self.cache.set("application_key_id", application_key_id)

    @_raise_missing_if_result_is_none
    def get_application_key(self):
        return self.cache.get("application_key")

    @_raise_missing_if_result_is_none
    def get_application_key_id(self):
        return self.cache.get("application_key_id")

    @_raise_missing_if_result_is_none
    def get_account_id(self):
        return self.cache.get("account_id")

    @_raise_missing_if_result_is_none
    def get_api_url(self):
        return self.cache.get("api_url")

    def get_account_auth_token(self):
        """Named different from cached value"""
        auth_token = self.cache.get("auth_token")
        if auth_token is None:
            raise MissingAccountData("auth_token")
        return auth_token

    @_raise_missing_if_result_is_none
    def get_download_url(self):
        return self.cache.get("download_url")

    @_raise_missing_if_result_is_none
    def get_realm(self):
        return self.cache.get("realm")

    @_raise_missing_if_result_is_none
    def get_absolute_minimum_part_size(self):
        return self.cache.get("absolute_minimum_part_size")

    @_raise_missing_if_result_is_none
    def get_recommended_part_size(self):
        return self.cache.get("recommended_part_size")

    @_raise_missing_if_result_is_none
    def get_allowed(self):
        return self.cache.get("allowed")

    def get_s3_api_url(self):
        return self.cache.get("s3_api_url") or ""

    def get_bucket_id_or_none_from_bucket_name(self, bucket_name: str) -> Optional[str]:
        try:
            return self.cache.get(_bucket_cachekey(bucket_name))
        except KeyError as e:
            logger.debug(f"cache miss {bucket_name}: {e}")
            return None

    def get_bucket_name_or_none_from_bucket_id(self, bucket_id: str) -> Optional[str]:
        try:
            for bucket_name in self.cache.get("bucket_names", []):
                bucket_id = self.cache.get(_bucket_cachekey(bucket_name))
                if bucket_id:
                    return bucket_name
            logger.debug(f"cache miss {bucket_id}")
        except KeyError as e:
            logger.debug(f"cache miss {bucket_id}: {e}")
        return None

    def refresh_entire_bucket_name_cache(self, name_id_iterable: Iterable[Tuple[str, str]]):
        bucket_names_to_remove = set(self.cache.get("bucket_names", [])) - {name for name, id in name_id_iterable}
        for (bucket_name, bucket_id) in name_id_iterable:
            self.cache.set(_bucket_cachekey(bucket_name), bucket_id)
        for bucket_name in bucket_names_to_remove:
            self.remove_bucket_name(bucket_name)

    def save_bucket(self, bucket: StoredBucketInfo):
        self.cache.set(_bucket_cachekey(bucket.name), bucket.id_)
        self.cache.set("bucket_names", list(self.cache.get("bucket_names", [])) + [bucket.name])

    def remove_bucket_name(self, bucket_name):
        self.cache.set("bucket_names", [n for n in self.cache.get("bucket_names", []) if n != bucket_name])
        self.cache.delete(_bucket_cachekey(bucket_name))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}{{cacheName={self._cacheName},cache={self.cache}}}"


def _bucket_cachekey(bucket_name: str) -> str:
    return hash(f"bucket-name__{bucket_name}".encode()).hexdigest()
